/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C) @BABOLO  2002..2010 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright Sergey Kosyakov ks@itp.ac.ru 1999\n\
@(#)Copyright (C) @BABOLO  2002..2010 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: tund.c,v 1.10 2013/10/14 17:58:30 babolo Exp $";
#endif /* not lint */

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <poll.h>
#include <time.h>
#include <mife.h>
#include <err.h>
#include <babolo/parser.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "tund.h"

int debug = 0;
int socket_control;
int control_port;
Seq *stun;

static void
usage(int ex) {
    fprintf( stderr
           , "tund " VERS "  " DATE " @BABOLO http://www.babolo.ru/\n"
             " -E - give arg of next flag from environment variable\n"
             " -q - quiet\n"
             " -r flags - recursive flags without variable substitution, for use in #! string\n"
             " -R flags - recursive flags with variable substitution, for use in #! string\n"
             " -v[v...] - be more verbose\n"
             " -? - short help\n"
           );
    exit(ex);
}

static void
proceed_tunnel(int tun_ind) {
    struct sockaddr_in saddr;
    Tunnel *tun, *t;
    int res;
    int sd;
    int i;

    tun = (Tunnel *)stun->buf[tun_ind];

    for (i = 0; i < tun_ind; i++) {
        t = (Tunnel *)stun->buf[i];
        if  (t->dp == tun->dp)
            Error("Illegal configuration: tunnels should have different divert sockets");
        if  (t->tunAddr == tun->tunAddr)
            Error("Illegal configuration: tunnels should have different labels");
    }
    bzero(&saddr, sizeof(saddr));
    saddr.sin_port = htons((unsigned short)tun->dp);
    saddr.sin_family = AF_INET;
    sd = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT);
    res = bind(sd, (struct sockaddr *)&saddr, sizeof(saddr));
    if  (res) Error("Can not bind to divert port %d\n", tun->dp);
    fcntl(sd, F_SETFL, FNDELAY);                                              /* Set to non-blocking */
    tun->socketDivert = sd;
#ifdef DEBUG
    if  (debug) Log( "Divert port %d is binded to socket %d for t=%d"
                   , tun->dp, tun->socketDivert, tun_ind
                   );
#endif /* DEBUG */
}

static void
do_cycle(tund_conf *cnf, struct pollfd *fds, int nfd, CBF *cb) {
    int ret;
    int i;

    while (1) {                                                                             /* Main loop */
        ret = poll(fds, nfd, 15000);
        if  (ret < 0) {
            if  (errno == EINTR) continue;
            Error("Poll returns error %d", ret);
        }
#ifdef DEBUG
        if  (debug) Log("Poll returns %d", ret); 
#endif /* DEBUG */
        for (i = 0; i < nfd; i++) {
            if  (fds[i].revents) {
#ifdef DEBUG
                if  (debug) Log("Event 0x%x on fd %d", fds[i].revents, fds[i].fd);
#endif /* DEBUG */
                cb[i].f(cnf, cb[i].tun_ind, fds[i].fd);
}   }   }   }

static int
tund(tund_conf *cnf) {
    ifBLIN_FLAG = cnf->flags;
    int ex = EX_OK;

    if  (!cnf->conf_nm) {
        if  (asprintf(&cnf->conf_nm, "%s/%s", cnf->confdir, CONFIG_FILE) < 0) {
            ifBLIN_QX0("no mem for asprintf conf_nm");
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
    }   }
    if  (!cnf->rsak_nm) {
        if  (asprintf(&cnf->rsak_nm, "%s/%s", cnf->confdir, RSAKEY_FILE) < 0) {
            ifBLIN_QX0("no mem for asprintf rsak_nm");
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
    }   }
    if  (!cnf->rsap_nm) {
        if  (asprintf(&cnf->rsap_nm, "%s/%s", cnf->confdir, RSAPUB_DB) < 0) {
            ifBLIN_QX0("no mem for asprintf rsap_nm");
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
    }   }
    if  (!cnf->seed_nm) {
        if  (asprintf(&cnf->seed_nm, "%s/%s", cnf->confdir, SEED_FILE) < 0) {
            ifBLIN_QX0("no mem for asprintf seed_nm");
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
    }   }
    {   struct addrinfo *hints, *r;

        if  (!(hints = calloc(1, sizeof(struct addrinfo)))) {
            ifBLIN_QX0("no mem for hints");
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
        }
        hints->ai_family = PF_INET;
        hints->ai_socktype = SOCK_DGRAM;
        hints->ai_protocol = IPPROTO_UDP;
        hints->ai_flags = AI_ADDRCONFIG | AI_PASSIVE;
        if  (getaddrinfo(cnf->bindaddr, cnf->port, hints, &r)) {
            ifBLIN_QX2( "getaddrinfo: No %s:%s"
                      , cnf->bindaddr ? cnf->bindaddr : ""
                      , cnf->port ? cnf->port : ""
                      );
            if  (!(cnf->flags & TUND_YPORT)) {
            if  (r) freeaddrinfo(r);
                cnf->port = "9001";
                if  (!getaddrinfo(cnf->bindaddr, cnf->port, hints, &r)) goto c1;
            }
            if  (r) freeaddrinfo(r);
            ifBLIN_QW0("getaddrinfo %s:%s", cnf->bindaddr, cnf->port);
            ex = EX_NOHOST;
            goto out;
        }
    c1:
        if  ((cnf->isock = socket(r->ai_family, r->ai_socktype, r->ai_protocol)) < 0) {
            /* There is list by r->ai_next pointer XXXX */
            ifBLIN_QW0( "IN socket %s:%s"
                      , cnf->bindaddr ? cnf->bindaddr : ""
                      , cnf->port ? cnf->port : ""
                      );
            ex = EX_NOHOST;
            goto out;
        }
        if  (bind(cnf->isock, r->ai_addr, r->ai_addrlen)) {
            ifBLIN_QW0( "IN bind %s:%s"
                      , cnf->bindaddr ? cnf->bindaddr : ""
                      , cnf->port ? cnf->port : ""
                      );
            ex = EX_NOHOST;
            goto out;
        }
        cnf->s_port = ((struct sockaddr_in *)r->ai_addr)->sin_port;
        cnf->s_addr = ((struct sockaddr_in *)r->ai_addr)->sin_addr.s_addr;
        if  (r) freeaddrinfo(r);
    }
    if  (fcntl(cnf->isock, F_SETFL, FNDELAY) == -1) {                                /* Non-blocking */
        ifBLIN_QW0( "IN fcntl FNDELAY %s:%s"
                  , cnf->bindaddr ? cnf->bindaddr : ""
                  , cnf->port ? cnf->port : ""
                  );
        ex = EX_NOHOST;
        goto out;
    }
    control_port = ntohs(cnf->s_port);
    ifBLIN_QX2( "IN socket %d = %u.%u.%u.%u:%u"
              , cnf->isock
              , cnf->s_addr & 0xFF, (cnf->s_addr >> 8) & 0xFF
              , (cnf->s_addr >> 16) & 0xFF, (cnf->s_addr >> 24) & 0xFF
              , ntohs(cnf->s_port)
              );
    cnf->stun = stun = seq_new(cnf);
    if  (parse_config(cnf)) {
        ifBLIN_QW0("parse_config");
        ex = EX_USAGE;
        goto out;
    }
    if  (init_rng(cnf)) {
        ifBLIN_QW0("init_rng");
        ex = EX_USAGE;
        goto out;
    }
    if  (init_secur(cnf)) {
        ifBLIN_QW0("init_secur");
        ex = EX_USAGE;
        goto out;
    }
    if  (init_alarm(cnf)) {
        ifBLIN_QW0("init_alarm");
        ex = EX_USAGE;
        goto out;
    }
    if  (!(cnf->flags & TUND_FOREG)) {
        if  (daemon(0, 0)) {
            ifBLIN_QW0("daemon");
            ex = EX_OSERR;
            goto out;
        }
        blin_ctl(BLIN_CTL_WHAT | BLIN_CTL_LEVM, BLIN_MODLOGR);
    }
    if  (unlink(cnf->pid_nm)) ifBLIN_QW1("unlink PID %s", cnf->pid_nm);
    if  ((cnf->pid_fd = open(cnf->pid_nm, O_CREAT | O_TRUNC | O_WRONLY | O_EXLOCK, 0666)) < 0) {
        ifBLIN_QW0("open PID %s", cnf->pid_nm);
        ex = EX_IOERR;
        goto out;
    }
    {   char *pid;
        int i;

        i = asprintf( &pid, "%d\n%u.%u.%u.%u:%u\n"
                    , getpid()
                    , cnf->s_addr & 0xFF, (cnf->s_addr >> 8) & 0xFF
                    , (cnf->s_addr >> 16) & 0xFF, (cnf->s_addr >> 24) & 0xFF
                    , ntohs(cnf->s_port)
                    );
        if  (!pid || (i < 0)) {
            ifBLIN_QW0("asprintf");
            ex = EX_IOERR;
            goto out;
        }
        if  (mife_writ(cnf->pid_fd, pid, 0) != i) {
            ifBLIN_QW0("mife_writ %s", cnf->pid_nm);
            ex = EX_IOERR;
            goto out;
        }
        free(pid);
    }
    {   int i;

        for (i = 0; i < stun->length; i++) proceed_tunnel(i);     /* Set up tunnel(s) divert sockets */
    }
    {   int i;
        struct pollfd *fds;
        Tunnel *tun;
        CBF *cb;
        int nfd;

        socket_control = cnf->isock;
        nfd = stun->length + 1;
        if  (!(fds = calloc(nfd, sizeof(struct pollfd)))) Error("Can not allocate memory for pollfd");
        if  (!(cb = calloc(nfd, sizeof(CBF)))) Error("Can not allocate memory for callbacks");
        for (i = 0; i < stun->length; i++) {
            tun = (Tunnel *)stun->buf[i];
            fds[i].fd = tun->socketDivert;
            fds[i].events = POLLRDNORM;
            cb[i].tun_ind = i;
            cb[i].f = input_divert;
            if  (!(tun->flags & TUNF_ENC)) {
                tun->state = TUNS_PRC;
            } else if (start_secur_tunnel(cnf, i)) {          /* Initiate encryption for that tunnel */
                ifBLIN_QW1("No start tunel %d", i);
        }   }
        fds[stun->length].fd = cnf->isock;
        fds[stun->length].events = POLLRDNORM;
        cb[stun->length].f = read_packet;
        cb[stun->length].tun_ind = -1;                                     /* Does have no any means */
        do_cycle(cnf, fds, nfd, cb);
    }
out:if  (!ex) ex = EX_SOFTWARE;
    flock(cnf->pid_fd, LOCK_UN);
    close(cnf->pid_fd);
    cnf->pid_fd = -1;
    return(ex);
}

int
main(int argc, char **argv) {
    babolo_opt *opt = NULL;
    int i, ex = EX_OK;
    u_int32_t flocal;
#   define EFLAGNOW  1
#   define EFLAGNEXT (EFLAGNOW << 1)
    tund_conf *cnf;

    if  (!(cnf = calloc(1, sizeof(tund_conf)))) {
        ifBLIN_QU0(-1, "no mem for tund_conf");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    cnf->pid_fd = -1;
    cnf->port = "tund";
    cnf->flags = BLIN_VER1;
    cnf->pid_nm = LOCK_FILE;
    cnf->confdir = CONF_DIR;
    if  (babolo_openopt(cnf->flags | Bpars_NONU, &opt, argc, argv, "b:dD:Ef:k:p:qr:R:s:u:v?")) {
        errx(EX_SOFTWARE, "babolo_openopt");
    }
    for (flocal = 0; (i = babolo_getopt(&opt)) != 0; flocal >>= 1) {
        babolo_parm *bp;
        char *o;

        switch (i) {
        case 'b': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; cnf->bindaddr = o;
                ; break;
        case 'd': debug = 1;
                ; cnf->flags |= TUND_FOREG;
                ; break;
        case 'D': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; cnf->confdir = o;
                ; break;
        case 'E': flocal |= EFLAGNEXT;
                ; break;
        case 'f': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; cnf->conf_nm = o;
                ; break;
        case 'k': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; cnf->rsak_nm = o;
                ; break;
        case 'p': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; cnf->port = o;
                ; cnf->flags |= TUND_YPORT;
                ; break;
        case 'q': cnf->flags &= ~BLIN_VERMASK;
                ; break;
        case 'r': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; bp = babolo_getparm( Bpars_NOEN | Bpars_NOAL | Bpars_NONU
                                     , (u_char**)&o
                                     , (u_char*)""
                                     , 0
                                     );
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'R': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 's': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; cnf->seed_nm = o;
                ; break;
        case 'u': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; cnf->rsap_nm = o;
                ; break;
        case 'v': BLIN_VERBOSE(cnf->flags);
                ; break;
        case '?': usage(EX_OK);
                ; break;
        default : usage(EX_USAGE);
    }   }
    blin_ctl( BLIN_CTL_WHAT
            | BLIN_CTL_LEV2 | BLIN_CTL_LEV3 | BLIN_CTL_LEV4
            | BLIN_CTL_LEV5 | BLIN_CTL_LEV6 | BLIN_CTL_LEV7
            , BLIN_MODPRNT | BLIN_MODEOAU
            );
    ex = tund(cnf);
out:
    exit(ex);
}
